/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse)
{
    std::shared_ptr<Query> query = std::make_shared<Query>();
    if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse))
    {
        // 处理表名
        query->tables = std::move(x->tabs);
        /** TODO: 检查表是否存在 */
        for (auto &tab_name : query->tables)
        {
            if (!sm_manager_->db_.is_table(tab_name))
            {
                throw TableNotFoundError(tab_name);
            }
        }
        // 处理target list，在target list中添加上表名，例如 a.id
        // 将x->cols的Col类型转化为query->cols的TabCol类型
        for (auto &sv_sel_col : x->cols)
        {
            if(sv_sel_col->agg_type!=ast::AGGTYPE_NONE)
                query->is_agg = true;
            //将聚集函数与封装成列，一起处理
            TabCol sel_col = {.tab_name = sv_sel_col->tab_name, .col_name = sv_sel_col->col_name, .agg_type = sv_sel_col->agg_type, .as_name = sv_sel_col->as_name};
            query->cols.push_back(sel_col);
        }

        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);
        // select * 的情况
        if (query->cols.empty())
        {
            // select all columns
            for (auto &col : all_cols)
            {
                TabCol sel_col = {.tab_name = col.tab_name, .col_name = col.name};
                query->cols.push_back(sel_col);
            }
        }
        // 判断sel_col是否在选择的表的所有列中
        else
        {
            // infer table name from column name
            for (auto &sel_col : query->cols)
            {
                sel_col = check_column(all_cols, sel_col); // 列元数据校验
            }
        }

        // 处理where条件
        get_clause(x->conds, query->conds);
        // WHERE 子句中不能用聚集函数作为条件表达式
        for(auto &con : x->conds)
        {
            if(con->lhs->agg_type != ast::AGGTYPE_NONE)
            {
                throw WhereAggError();
            }
        }
        check_clause(query->tables, query->conds);

        // 处理分组
        if (!x->group_by_cols.empty())
        {
            for (auto &group_by_col : x->group_by_cols)
            {
                TabCol group_col = {.tab_name = group_by_col->tab_name, .col_name = group_by_col->col_name};
                query->group_by_cols.push_back(group_col);
            }
            // SELECT 列表中不能出现没有在 GROUP BY 子句中的非聚集列
            for (auto &sel_col : query->cols)
            {
                //检查query->cols在group_by_cols_colmeta中
                check_column_groupby(all_cols, query->group_by_cols, sel_col);
            }
            
            // std::vector<ColMeta> group_by_cols_colmeta;
            // get_cols(query->group_by_cols, group_by_cols_colmeta);
            // for (auto &sel_col : query->cols)
            // {
            //     //检查query->cols在group_by_cols_colmeta中
            //     check_column(group_by_cols_colmeta, sel_col); 
            // }
        }

        // 处理having
        // 怎么处理？？？
        // 1.聚集函数括号内的列在表中
        // 2.条件的右式是数字
        get_clause(x->having_conds, query->having_conds);
        // check_having_clause(query);


    }
    else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse))
    {
        /** TODO: */
        // 处理where条件
        get_clause(x->conds, query->conds);
        check_clause({x->tab_name}, query->conds);
        // 处理update的set值
        for (auto &sv_set : x->set_clauses)
        {
            SetClause set_clause;
            set_clause.lhs.col_name = sv_set->col_name;
            set_clause.lhs.tab_name = x->tab_name;
            set_clause.rhs = convert_sv_value(sv_set->val);

            TabMeta &lhs_tab = sm_manager_->db_.get_table(set_clause.lhs.tab_name);
            auto lhs_col = lhs_tab.get_col(set_clause.lhs.col_name);
            if (lhs_col->type == TYPE_FLOAT && set_clause.rhs.type == TYPE_INT) {
                set_clause.rhs.type = TYPE_FLOAT;
                set_clause.rhs.set_float(set_clause.rhs.int_val);
            }
            set_clause.rhs.init_raw(lhs_col->len);
            if(lhs_col->type!=set_clause.rhs.type){
                throw IncompatibleTypeError(coltype2str(lhs_col->type),coltype2str(set_clause.rhs.type));
            }
            query->set_clauses.push_back(set_clause);
        }
    }
    else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse))
    {
        // 处理where条件
        get_clause(x->conds, query->conds);
        check_clause({x->tab_name}, query->conds);
    }
    else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse))
    {
        // 处理insert 的values值
        for (auto &sv_val : x->vals)
        {
            query->values.push_back(convert_sv_value(sv_val));
        }
    }
    else
    {
        // do nothing
    }
    query->parse = std::move(parse);
    return query;
}

// 检查target是否在all_cols中
TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols, TabCol target)
{
    if (target.agg_type == ast::AGGTYPE_COUNT && target.col_name == "")
    {
        return target;
    }

    if (target.tab_name.empty())
    {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto &col : all_cols)
        {
            if (col.name == target.col_name)
            {
                if (!tab_name.empty())
                {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty())
        {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    }
    else
    {
        /** TODO: Make sure target column exists */
        bool flag = false;
        for (auto &col : all_cols)
        {
            if (col.name == target.col_name && col.tab_name == target.tab_name)
            {
                flag = true;
                break;
            }
        }
        if (!flag)
        {
            throw ColumnNotFoundError(target.col_name);
        }
    }
    return target;
}

// 检查target是否在groupby_cols中
// 若非，报错“failure”
// 用于检查SELECT 列表中不能出现没有在 GROUP BY 子句中的非聚集列
void Analyze::check_column_groupby(const std::vector<ColMeta> &all_cols, std::vector<TabCol> &groupby_cols, TabCol target)
{
    if (target.agg_type != ast::AGGTYPE_NONE)
    {
        return;
    }
    // 为groupby的列加上表名
    for (auto &groupby_col : groupby_cols)
        {
            if (groupby_col.tab_name.empty())
            {
                // Table name not specified, infer table name from column name
                std::string tab_name;
                for (auto &col : all_cols)
                {
                    if (col.name == target.col_name)
                    {
                        if (!tab_name.empty())
                        {
                            throw AmbiguousColumnError(groupby_col.col_name);
                        }
                        tab_name = col.tab_name;
                    }
                }
                if (tab_name.empty())
                {
                    throw ColumnNotFoundError(groupby_col.col_name);
                }
                groupby_col.tab_name = tab_name;
            }
        }
    // 为target的列加上表名
    if (target.tab_name.empty())
    {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto &col : all_cols)
        {
            if (col.name == target.col_name)
            {
                if (!tab_name.empty())
                {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty())
        {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    }

    /** TODO: Make sure target column exists */
    bool flag = false;
    for (auto &groupby_col : groupby_cols)
    {
        if (groupby_col.col_name == target.col_name && groupby_col.tab_name == target.tab_name)
        {
            flag = true;
            break;
        }
    }
    if (!flag)
    {
        throw SelectError();
    }
    return;
}


// 根据表名获取表中的所有列
// 用vector<ColMeta>传出结果
void Analyze::get_all_cols(const std::vector<std::string> &tab_names, std::vector<ColMeta> &all_cols)
{
    for (auto &sel_tab_name : tab_names)
    {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
}

// 增添
// 根据TabCol类型的列获取ColMeta类型的列
// 用vector<ColMeta>传出结果
void Analyze::get_cols(const std::vector<TabCol> &col, std::vector<ColMeta> &get_cols)
{
    for (auto &sel_col : col)
    {
        auto &sel_tab = sm_manager_->db_.get_table(sel_col.tab_name);
        const auto &sel_cols = sel_tab.get_col(sel_col.col_name);
        get_cols.insert(get_cols.end(), *sel_cols);
    }
}

// 分解where中的条件
void Analyze::get_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds, std::vector<Condition> &conds)
{
    // 怎么保证左边一定是列，是否要考虑类似"4<id"的条件？
    // 测试样例中没有出现
    conds.clear();
    for (auto &expr : sv_conds)
    {
        Condition cond;
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name, .agg_type = expr->lhs->agg_type, .as_name = expr->lhs->as_name};
        cond.op = convert_sv_comp_op(expr->op);
        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs))
        {
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        }
        else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs))
        {
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
        }
        conds.push_back(cond);
    }
}

// 检查where中的条件是否合法
void Analyze::check_clause(const std::vector<std::string> &tab_names, std::vector<Condition> &conds)
{
    // auto all_cols = get_all_cols(tab_names);
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto &cond : conds)
    {
        // Infer table name from column name
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if (!cond.is_rhs_val)
        {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }
        TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;
        if (cond.is_rhs_val)
        {
            if (lhs_type == TYPE_FLOAT && cond.rhs_val.type == TYPE_INT) {
                cond.rhs_val.type = TYPE_FLOAT;
                cond.rhs_val.set_float(cond.rhs_val.int_val);
            }
            cond.rhs_val.init_raw(lhs_col->len);
            rhs_type = cond.rhs_val.type;
        }
        else
        {
            TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        }
        if (lhs_type != rhs_type)
        {
            throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
        }
    }
}

Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val)
{
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val))
    {
        val.set_int(int_lit->val);
    }
    else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val))
    {
        val.set_float(float_lit->val);
    }
    else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val))
    {
        val.set_str(str_lit->val);
    }
    else
    {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op)
{
    std::map<ast::SvCompOp, CompOp> m = {
        {ast::SV_OP_EQ, OP_EQ},
        {ast::SV_OP_NE, OP_NE},
        {ast::SV_OP_LT, OP_LT},
        {ast::SV_OP_GT, OP_GT},
        {ast::SV_OP_LE, OP_LE},
        {ast::SV_OP_GE, OP_GE},
    };
    return m.at(op);
}
