/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include "parser/ast.h"
#include "limits.h"

class AggregateExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;  // 投影节点的儿子节点
    std::vector<TabCol> agg_cols_;
    std::vector<ColMeta> cols_;
    size_t len_;
    std::vector<Value> values_;
    std::vector<size_t> idxs_;

    // ast::AggType aggType_;//聚合类型
    // std::string asName_;//聚合出来的数据名称
    // long long num;
    // long long intRecord;
    // float floatRecord;
    // std::string charRecord;

   public:
    AggregateExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol> &agg_cols//, ast::AggType aggType, std::string asName
        )
    {
        prev_ = std::move(prev);
        agg_cols_ = std::move(agg_cols);
        size_t curr_offset = 0;
        auto &prev_cols = prev_->cols();
        for (auto &agg_col : agg_cols)
        {
            ColMeta col;
            struct Value value;
            if (agg_col.agg_type == ast::AGGTYPE_COUNT) {
                col.type = TYPE_INT;
                col.len = sizeof(int);
                value.set_int(0);
            }
            else {
                auto pos = get_col(prev_cols, agg_col);
                col.type = pos->type;
                col.len = pos->len;
                idxs_.push_back(pos - prev_cols.begin());
            }
            col.offset = curr_offset;
            curr_offset += col.len;
            cols_.push_back(col);
            value.type = col.type;
            values_.push_back(value);
        }
        len_ = curr_offset;

        // aggType_ = std::move(aggType);
        // asName_ = std::move(asName);
        // if (aggType_ != ast::AGGTYPE_COUNT) { //验证一下，如果是MAX,MIN,SUM的话只能有一列
        //     assert(prev_->cols().size() == 1);//这里假设了聚合的子节点一定是projection
        // }
    }
    std::unique_ptr<RmRecord> Next() override
    {
        bool init_ = true; // true的时候初始化
        std::unique_ptr<RmRecord> agg_record = std::make_unique<RmRecord>(len_);
        auto &prev_cols = prev_->cols();
        for (prev_->beginTuple(); !prev_->is_end(); prev_->nextTuple())
        {
            auto prev_record = prev_->Next();
            const char *prev_data = prev_record->data;
            for (size_t i = 0; i < cols_.size(); i++)
            {
                size_t j = 0;
                if (agg_cols_[i].agg_type==ast::AGGTYPE_COUNT)
                {
                    values_[i].int_val++;
                }
                else
                {
                    auto pos = prev_cols[idxs_[j]];
                    if (cols_[i].type==TYPE_INT)
                    {
                        int temp = *(int *)(prev_data + pos.offset);
                        if (init_)
                        {
                            values_[i].set_int(temp);
                            init_ = !init_;
                        }
                        else if (agg_cols_[i].agg_type==ast::AGGTYPE_MAX)
                        {
                            values_[i].int_val = std::max(values_[i].int_val, temp);
                        }
                        else if (agg_cols_[i].agg_type==ast::AGGTYPE_MIN)
                        {
                            values_[i].int_val = std::min(values_[i].int_val, temp);
                        }
                        else if (agg_cols_[i].agg_type==ast::AGGTYPE_SUM)
                        {
                            values_[i].int_val = values_[i].int_val + temp;
                        }
                    }
                    else
                    {
                        float temp = *(float *)(prev_data + pos.offset);
                        if (init_)
                        {
                            values_[i].set_float(temp);
                            init_ = !init_;
                        }
                        else if (agg_cols_[i].agg_type==ast::AGGTYPE_MAX)
                        {
                            values_[i].float_val = std::max(values_[i].float_val, temp);
                        }
                        else if (agg_cols_[i].agg_type==ast::AGGTYPE_MIN)
                        {
                            values_[i].float_val = std::min(values_[i].float_val, temp);
                        }
                        else if (agg_cols_[i].agg_type==ast::AGGTYPE_SUM)
                        {
                            values_[i].float_val = values_[i].float_val + temp;
                        }
                    }
                    j++;
                }
            }
        }
        char *agg_data = agg_record->data;
        for (size_t i = 0; i < cols_.size(); i++)
        {
            auto &col = cols_[i];
            if (values_[i].type=TYPE_INT)
            {
                memcpy(agg_data + col.offset, &(values_[i].int_val), sizeof(int));
            }
            else if (values_[i].type=TYPE_FLOAT)
            {
                memcpy(agg_data + col.offset, &(values_[i].float_val), sizeof(float));
            }
        }
        return agg_record;
    }
    // std::unique_ptr<RmRecord> Next() override {
    //     if (prev_->cols()[0].type == TYPE_INT) { //int型数据
    //         long long temp;
    //         if (aggType_ == ast::AGGTYPE_COUNT) {
    //             temp = 0;
    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 temp++;
    //                 auto Tuple = prev_->Next();
    //                 prev_->nextTuple();
    //             }

    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;

    //         } else if (aggType_ == ast::AGGTYPE_MAX) {
    //             bool init_ = true; //true的时候初始化

    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 int data_ = *(int *)rec_buf;
    //                 if (init_) {
    //                     temp = (long long)data_;
    //                     init_ = !init_;
    //                 } else {
    //                     temp = std::max(temp, (long long)data_);
    //                 }
    //                 prev_->nextTuple();
    //             }

    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;

    //         } else if (aggType_ == ast::AGGTYPE_MIN) {
    //             bool init_ = true; //true的时候初始化

    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 int data_ = *(int *)rec_buf;
    //                 if (init_) {
    //                     temp = (long long)data_;
    //                     init_ = !init_;
    //                 } else {
    //                     temp = std::min(temp, (long long)data_);
    //                 }
    //                 prev_->nextTuple();
    //             }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         } else if (aggType_ == ast::AGGTYPE_SUM) {
    //             temp = 0;
    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 int data_ = *(int *)rec_buf;
    //                 temp += (long long)data_;
    //                 prev_->nextTuple();
    //                 }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         }
    //     } else if (prev_->cols()[0].type == TYPE_FLOAT) {
    //         float temp;
    //         if (aggType_ == ast::AGGTYPE_COUNT) {
    //             int temp = 0;
    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 temp++;
    //                 auto Tuple = prev_->Next();
    //                 prev_->nextTuple();
    //             }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         } else if (aggType_ == ast::AGGTYPE_MAX) {
    //             bool init_ = true; //true的时候初始化

    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 float data_ = *(float *)rec_buf;
    //                 if (init_) {
    //                     temp = data_;
    //                     init_ = !init_;
    //                 } else {
    //                     temp = std::max(temp, data_);
    //                 }
    //                 prev_->nextTuple();
    //             }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         } else if (aggType_ == ast::AGGTYPE_MIN) {
    //             bool init_ = true; //true的时候初始化

    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 float data_ = *(float *)rec_buf;
    //                 if (init_) {
    //                     temp = data_;
    //                     init_ = !init_;
    //                 } else {
    //                     temp = std::min(temp, data_);
    //                 }
    //                 prev_->nextTuple();
    //             }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         } else if (aggType_ == ast::AGGTYPE_SUM) {
    //             temp = 0;
    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 float data_ = *(float *)rec_buf;
    //                 temp += data_;
    //                 prev_->nextTuple();
    //                 }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         }
    //     } else if (prev_->cols()[0].type == TYPE_STRING) {
    //         std::string temp;
    //         if (aggType_ == ast::AGGTYPE_COUNT) {
    //             int temp = 0;
    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 temp++;
    //                 prev_->nextTuple();
                    
    //             }
                
    //             std::string output = std::to_string(temp);
    //             auto rec = std::make_unique<RmRecord>(output.length());
    //             memcpy(rec->data, output.c_str(), output.length());

    //             return rec;
                
    //         } else if (aggType_ == ast::AGGTYPE_MAX) {
    //             bool init_ = true; //true的时候初始化

    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 std::string data_ = std::string((char *)rec_buf, col.len);
    //                 if (init_) {
    //                     temp = data_;
    //                     init_ = !init_;
    //                 } else {
    //                     int val = temp.compare(data_);
    //                     if (val<0) {
    //                         temp = data_;
    //                     }
    //                 }
    //                 prev_->nextTuple();
    //             }
                
    //             auto rec = std::make_unique<RmRecord>(temp.length());
    //             memcpy(rec->data, temp.c_str(), temp.length());

    //             return rec;
                
    //         } else if (aggType_ == ast::AGGTYPE_MIN) {
    //             bool init_ = true; //true的时候初始化

    //             prev_->beginTuple();
    //             while(!prev_->is_end()) {
    //                 auto &col = prev_->cols()[0];
    //                 auto Tuple = prev_->Next();
    //                 char *rec_buf = Tuple->data + col.offset;
    //                 std::string data_ = std::string((char *)rec_buf, col.len);
    //                 if (init_) {
    //                     temp = data_;
    //                     init_ = !init_;
    //                 } else {
    //                     int val = temp.compare(data_);
    //                     if (val>0) {
    //                         temp = data_;
    //                     }
    //                 }
    //                 prev_->nextTuple();
    //             }
                
    //             auto rec = std::make_unique<RmRecord>(temp.length());
    //             memcpy(rec->data, temp.c_str(), temp.length());

    //             return rec;
                
    //         }
    //     }
    // }

    // void beginTuple() override { prev_->beginTuple(); };
    
    // void nextTuple() override {
    //     assert(!prev_->is_end());
    //     prev_->nextTuple();
    // }

    // bool is_end() const override { return true; } 

    const std::vector<ColMeta> &cols() const override { return cols_; };
    Rid &rid() override { return _abstract_rid; }

    
};