/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    bool isend;
    bool isfirst;

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);
        isfirst=true;
    }
    size_t tupleLen() const override { return len_; }
    void beginTuple() override {
        bool find=false;
        auto record = std::make_unique<RmRecord>(len_);
        for(left_->beginTuple(); !left_->is_end()&&!find;){
            auto left_rec=left_->Next();
            for(right_->beginTuple(); !right_->is_end(); right_->nextTuple()){
                auto right_rec=right_->Next();
                memcpy(record->data, left_rec->data, left_rec->size);
                memcpy(record->data + left_rec->size, right_rec->data, right_rec->size);
                if(eval_conds(cols_,fed_conds_,record.get())){
                    find=true;
                    break;
                }
            }
            if(!find){
                left_->nextTuple();
            }
        }
    }

    void nextTuple() override {
        bool find=false;
        auto record = std::make_unique<RmRecord>(len_);
        while(!left_->is_end()&&!find){
            auto left_rec=left_->Next();
            // if(isfirst){right_->nextTuple();}else{right_->beginTuple();}
            if(isfirst){right_->nextTuple();}
            for(; !right_->is_end(); right_->nextTuple()){
                auto right_rec=right_->Next();
                memcpy(record->data, left_rec->data, left_rec->size);
                memcpy(record->data + left_rec->size, right_rec->data, right_rec->size);
                if(eval_conds(cols_,fed_conds_,record.get())){
                    find=true;
                    isfirst=true;
                    break;
                }
            }
            if(!find){
                 left_->nextTuple();
                 right_->beginTuple();
                 isfirst=false;
            }
        }
        if(!find){isend=true;}
    }

    std::unique_ptr<RmRecord> Next() override {
        auto record = std::make_unique<RmRecord>(len_);
        auto left_rec=left_->Next();
        auto right_rec=right_->Next();
        memcpy(record->data, left_rec->data, left_rec->size);
        memcpy(record->data + left_rec->size, right_rec->data, right_rec->size);
        return record;
    }

    bool is_end() const override { return isend; }

    // const std::vector<ColMeta> &cols() const override { return cols_; }

    Rid &rid() override { return _abstract_rid; }

    bool eval_cond(const std::vector<ColMeta> &rec_cols, const Condition &cond, const RmRecord *rec) {
        // 获取左操作数
        auto lhs_col = get_col(rec_cols, cond.lhs_col);
        char *lhs = rec->data + lhs_col->offset;

        // 获取右操作数
        char *rhs;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            //右操作数是值
            rhs_type = cond.rhs_val.type;
            rhs = cond.rhs_val.raw->data;
        } else {
            // 右操作数是列
            auto rhs_col = get_col(rec_cols, cond.rhs_col);
            rhs_type = rhs_col->type;
            rhs = rec->data + rhs_col->offset;
        }

        // 断言左右操作数的类型相同
        assert(rhs_type == lhs_col->type);

        // 比较操作
        int cmp = ix_compare(lhs, rhs, rhs_type, lhs_col->len);

        // 根据操作符进行条件判断
        switch (cond.op) {
            case OP_EQ:
                return cmp == 0;
            case OP_NE:
                return cmp != 0;
            case OP_LT:
                return cmp < 0;
            case OP_GT:
                return cmp > 0;
            case OP_LE:
                return cmp <= 0;
            case OP_GE:
                return cmp >= 0;
            default:
                throw InternalError("Unexpected op type");
        }
    }
    const std::vector<ColMeta> &cols() const override { return cols_; }

    //这段代码的作用是在给定的记录 rec 中，评估一组条件 conds 是否全部成立。它利用了 C++ STL 中的 std::all_of 算法来遍历并评估每个条件，并返回整体的评估结果。
    bool eval_conds(const std::vector<ColMeta> &rec_cols, const std::vector<Condition> &conds, const RmRecord *rec) {
        return std::all_of(conds.begin(), conds.end(),
                           [&](const Condition &cond) { return eval_cond(rec_cols, cond, rec); });
    }
};